# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#           Resolution d'equations            #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# Le but de cette librairie est de permettre  terme la rsolution de toute quation, inquation,
# ou combinaison logique d'quation et d'inquations linaires ou de degr 2  une variable, dans R.
# (Dans un second temps, ceci pourra tre tendu aux valeurs absolues, quations trigonomtriques les plus simples,...)

# L'utilisateur doit n'avoir  faire qu' une unique fonction, "resoudre()".
# Eventuellement, il pourra prciser l'ensemble o rsoudre, qui devra tre une runion d'intervalle.
# (En particulier, il n'est pas question de rsoudre dans Z par exemple).

# version unicode


from intervalles import *

class Systeme:
    u"""Un systme d'quations et/ou d'inquations  une variable.
    
    La reprsentation d'une (in)quation en X est base sur la classe Polynome.
    
    Reprsentation interne :
    chaque (in)quation correspond  un couple (Polynome, int).
    Si le polynome est P, n l'entier, l'(in)quation reprsente est :
    P(X) R 0
    avec la table de correspondance suivante :
    n  |  R
    -------
    1     =
    -1   <>
    2     >
    -2   <=
    3    >=
    -3    <
    
    Les systemes sont traduits par une liste de listes de couples (sic!)
    Chacune de listes de couples de la liste principale reprsente un systme en "et".
    Si c1, c2, c3, c4 et c5 sont des couples reprsentant chacun une (in)quation:
    [[c1, c2, c3], [c4, c5]] signifie : (c1 et c2 et c3) ou (c4 et c5).
    
    Exemple:
    '(X+3<0 et 2X+2>0) ou (X>3 et X<=7) ou (X=2)'
    -> [[(X+3, -3), (2X+2, 2)], [(X-3, 2), (X-7, -2)], [(X-2, 1)]]
    """
    
    def __init__(self, systeme):
        self._systeme =  systeme
        
        
    def __or__(self, y):
        if y is True:
            return True
        return Systeme(self._systeme + y._systeme)
        
    def __ror__(self, y):
        return self.__or__(y)
        
    def __and__(self, y):
        if y is False:
            return self
        return Systeme([condition_i + condition_j for condition_i in self._systeme for condition_j in y._systeme])
        
    def __rand__(self, y):
        return self.__or__(y)

    def __invert__(self):
        return Systeme([[self._nier(condition) for condition in liste] for liste in self._systeme])
        
    def _nier(self, condition):
        u"Nie une condition (ie. un couple polynome-relation)."
        return condition[0], -condition[1]
        

def _eq2interv(polynome, relation):
    p = polynome
    r = abs(relation)
    d = p.deg()
    a = p.coeffs[0]
    I = Vide()
    if d == 1:  # premier degr
        sol = p.racines()[0]
        if r&1: # =
            I += Singleton(sol)
        elif r&2: # >
            if a > 0:
                I += Intervalle(sol, infini, inf_inclus=False)
            else:
                I += Intervalle(-infini, sol, sup_inclus=False)
    
    elif d == 2:    # deuxime degr
        sols = p.racines()
        sols.sort()
        de = p.delta()
        if r&1: # =
            I += Ensemble_discret(*sols)
        if r&2: # >
            if a > 0:
                if de < 0:
                    I += Intervalle() # R tout entier
                elif de == 0:
                    I += -Singleton(sols[0])
                else:
                    I += Intervalle(-infini, sols[0], sup_inclus=False) + Intervalle(sols[1], infini, inf_inclus=False)
            else:
                if de > 0:
                    I += Intervalle(sols[0], sols[1], inf_inclus=False, sup_inclus=False)
    
    if relation < 0:
        I = - I # son complementaire
    return I
    
def _somme(*liste):
    return reduce(lambda x,y:x+y, liste, Vide())

def _produit(*liste):
    return reduce(lambda x,y:x*y, liste, Intervalle())


def resoudre_systeme(systeme, ensemble = Intervalle()):
    if systeme is True:
        ensemble_des_solutions = Intervalle()
    elif systeme is False:
        ensemble_des_solutions = Vide()
    else:
        ensemble_des_solutions = _somme(*(_produit(*(_eq2interv(*condition) for condition in liste)) for liste in systeme._systeme))
    return ensemble_des_solutions*ensemble